package Apache::Bluestem;
use strict;
use warnings;

# This setting is so that the lack of an allowed.netids file in a
# directory will be remembered for 10 seconds per apache process.
use constant ALLOWED_NETIDS_CACHE_TIMEOUT => 10;

use Socket qw(inet_ntoa inet_aton sockaddr_in unpack_sockaddr_in AF_INET);
use Fcntl qw(:mode);
use FileHandle ();
use CGI::Cookie ();
use Digest::HMAC_SHA1 ();
use URI::Escape qw(uri_escape);
use Bluestem::Client ();
use MIME::Base64 qw(encode_base64);
my %bsc_cache = ();
my %no_allowed_netids_cache = ();
my $have_authtkt;

BEGIN {
	my $loaded = 0;
	my $error = '';
	eval q/
		# For old mod_perl
		use Apache ();
		use Apache::Constants qw(OK REDIRECT FORBIDDEN DECLINED);
	
		sub modperl_version{ "1.x" };
		sub r_add_err_header { $_[0]->err_header_out($_[1],$_[2]) }
		sub r_document_root  { $_[0]->document_root() }
		sub r_header_in      { $_[0]->header_in($_[1]) }
		sub r_set_header_in  { $_[0]->header_in($_[1],$_[2]) }
		sub r_local_port     { my($port) = unpack_sockaddr_in($_[0]->connection->local_addr); $port }
		sub r_path           { $_[0]->uri() }
		sub r_perlvar        { $_[0]->dir_config( $_[1] ) }
		sub r_query_string   { $_[0]->args() }
		sub r_remote_ip      { $_[0]->connection->remote_ip }
		sub r_remote_host    { $_[0]->connection->remote_host }
		sub r_server_name    { $_[0]->server->server_hostname() }
		sub r_user           { $_[0]->user($_[1]) }
		sub r_subprocess_env { $_[0]->subprocess_env($_[1],$_[2]) }

		$loaded = 1;
	/;
	$error .= "Failed to load for mod_perl 1.x:\n$@"
		unless $loaded;

	eval q/
		# For RHEL4 mod_perl-1.99
		use APR::Table ();
		use APR::SockAddr ();
		use Apache::RequestRec ();
		use Apache::ServerRec ();
		use Apache::RequestUtil ();
		use Apache::Response ();
		use Apache::Const -compile =>qw(OK REDIRECT FORBIDDEN DECLINED);
		use Apache::Connection ();

		sub modperl_version{ "1.99" };
		sub REDIRECT { Apache::REDIRECT };
		sub OK { Apache::OK };
		sub FORBIDDEN { Apache::FORBIDDEN };
		sub DECLINED { Apache::DECLINED };
		sub r_add_err_header { $_[0]->err_headers_out->add( $_[1] => $_[2] ) }
		sub r_document_root  { $_[0]->document_root() }
		sub r_header_in      { $_[0]->headers_in->{$_[1]} || '' }
		sub r_set_header_in  { $_[0]->headers_in->set( $_[1] => $_[2] ) }
		sub r_local_port     { $_[0]->connection->local_addr->port }
		sub r_path           { $_[0]->uri() }
		sub r_perlvar        { $_[0]->dir_config( $_[1] ) }
		sub r_query_string   { $_[0]->args() }
		sub r_remote_ip      { $_[0]->connection->remote_ip }
		sub r_remote_host    { $_[0]->connection->remote_host }
		sub r_server_name    { $_[0]->server->server_hostname() }
		sub r_user           { $_[0]->user($_[1]) }
		sub r_subprocess_env { $_[0]->subprocess_env($_[1],$_[2]) }

		$loaded = 1;
	/ unless $loaded;
	$error .= "Failed to load for mod_perl 1.99:\n$@"
		unless $loaded;

	eval q/
		# For mod_perl-2.x
		use APR::Table ();
		use APR::SockAddr ();
		use Apache2::RequestRec ();
		use Apache2::ServerRec ();
		use Apache2::RequestUtil ();
		use Apache2::Response ();
		use Apache2::Const -compile =>qw(OK REDIRECT FORBIDDEN DECLINED);
		use Apache2::Connection ();
		
		sub modperl_version{ "2.x" };
		sub REDIRECT { Apache2::Const::REDIRECT };
		sub OK { Apache2::Const::OK };
		sub FORBIDDEN { Apache2::Const::FORBIDDEN };
		sub DECLINED { Apache2::Const::DECLINED };
		sub r_add_err_header { $_[0]->err_headers_out->add( $_[1] => $_[2] ) }
		sub r_document_root  { $_[0]->document_root() }
		sub r_header_in      { $_[0]->headers_in->{$_[1]} || '' }
		sub r_set_header_in  { $_[0]->headers_in->set( $_[1] => $_[2] ) }
		sub r_local_port     { $_[0]->connection->local_addr->port }
		sub r_path           { $_[0]->uri() }
		sub r_perlvar        { $_[0]->dir_config( $_[1] ) }
		sub r_query_string   { $_[0]->args() }
		sub r_remote_ip {
			my $c = $_[0]->connection;
			$c->can('client_ip') ? $c->client_ip : $c->remote_ip;
		}
		sub r_remote_host {
			my $c = $_[0]->connection;
			$c->can('client_host') ? $c->client_host : $c->remote_host;
		}
		sub r_server_name    { $_[0]->server->server_hostname() }
		sub r_user           { $_[0]->user($_[1]) }
		sub r_subprocess_env { $_[0]->subprocess_env($_[1],$_[2]) }

		$loaded = 1;
	/ unless $loaded;
	$error .= "Failed to load for mod_perl 2.x:\n$@"
		unless $loaded;

	eval q/
		use Apache::AuthTkt ();
		$have_authtkt = 1;
	/;

	$loaded or die $error;
}

sub read_session_cookies ($$)
{
	my($r,$bsc) = @_;
	my %cookies = CGI::Cookie->parse(r_header_in($r,'Cookie'));
	my $cookie = $cookies{$bsc->{conf}{client_cookiename}};
	my($cacheid,$clientkey);
	if( $cookie and $cookie->value() ) {
		$cookie = $cookie->value();
		$cookie =~ s/^"|"$//g;
		($cacheid,$clientkey) = split( /~/, $cookie );
	}
	my %paservers = ();
	for my $idserver ( keys %{$bsc->{conf}{id_servers}} ) {
		my $cookiename = "Bluestem_$idserver";
		$cookiename =~ tr/:/_/;
		next unless $cookies{$cookiename};
		$paservers{$idserver} = $cookies{$cookiename}->value();
	}
	return($cacheid,$clientkey,\%paservers);
}

sub __allowed_netids_getline ($)
{
	my $fh = shift;
	while(1) {
		my $line = <$fh>;
		$line or return;
		chomp $line;
		$line =~ s/#.*//;
		$line =~ s/^\s+|\s+$//g;
		return $line if $line;
	};
}

sub __parse_allowed_netids ($$)
{
	my($fh,$bsc) = @_;

	# Parse allowed.netids
	my $contact;
	my @allow_ip;
	my @allow_host;
	my %allow_netid;
	my %allow_group;
	my $line = __allowed_netids_getline($fh);
	SECTION: while( $line ) {
		if( $line eq '<contact>' ) {
			# FIXME - Make parser smarter here?
			$contact = __allowed_netids_getline($fh);
		} elsif( $line =~ m/<allow ips?>/ ) {
			$line = __allowed_netids_getline($fh);
			while( $line ) {
				next SECTION if $line =~ m/^</;
				if( $line =~ m/^\d/ ) {
					push @allow_ip,$line;
				} else {
					push @allow_host,$line;
				}
				$line = __allowed_netids_getline($fh);
			}
		} elsif( $line =~ m/<allow netids?(?:\s+domain="([a-zA-Z.-]+)"\s*)?>/ ) {
			my $domain = ( $1 and $1 ne $bsc->{conf}{default_domain} ) ? $1 : '';
			$line = __allowed_netids_getline($fh);
			while( $line ) {
				next SECTION if $line =~ m/^</;
				if( $line =~ s/@(.*)// ){
					my $d = $1;
					$d = ''
						if $d eq $bsc->{conf}{default_domain};
					$d = '@'.$d if $d;
					$allow_netid{$line.$d} = 1;
				} elsif( $domain ){
					$allow_netid{$line.'@'.$domain} = 1;
				} else {
					$allow_netid{$line} = 1;
				}
				$line = __allowed_netids_getline($fh);
			}
		} elsif( $line =~ m/<allow groups?\s*>/ ) {
			$line = __allowed_netids_getline($fh);
			while( $line ) {
				next SECTION if $line =~ m/^</;
				if( $line =~ s/@(.*)// ) {
					my $d = $1 || '';
					$d = ($d eq $bsc->{conf}{default_domain}) ? '' : $d;
					$d = '@'.$d if $d;
					$allow_group{$line.$d} = 1;
				} else {
					$allow_group{$line} = 1;
				}
				$line = __allowed_netids_getline($fh);
			}
		} else {
			if( $line =~ s/@(.*)// ){
				my $d = $1;
				$d = ''
					if $d eq $bsc->{conf}{default_domain};
				$d = '@'.$d if $d;
				$allow_netid{$line.$d} = 1;
			} else {
				$allow_netid{$line} = 1;
			}

		}
		$line = __allowed_netids_getline($fh);
	}
	return {
		contact => $contact,
		ip => \@allow_ip,
		host => \@allow_host,
		netid => \%allow_netid,
		group => \%allow_group,
	};
}

sub __get_allowed_netids ($$)
{
	my($r,$bsc) = @_;
	my $docroot = r_document_root($r);
	my $path = r_path($r);
	$path =~ s{/+}{/}g;

	# Default searchpath
	my $basedir = $docroot;
	my $searchpath = $path;

	# Check for homedir path
	if( $bsc->{conf}{user_public_html} and $path =~ m{^/~([^/]+)(.*)} ) {
		my($user,$path) = ($1,$2);
		my @pwent = getpwnam($user);
		if( @pwent ) {
			my $homedir = $pwent[7];
			my $wwwdir = $bsc->{conf}{user_public_html};
                        if( $wwwdir =~ m/%{/ ) {
                                $basedir = $wwwdir;
                                $basedir =~ s/%{home}/$homedir/g;
                                $basedir =~ s/%{user}/$user/g;
                        } else {
                                $basedir = "$homedir/$wwwdir";
                        }
                        $searchpath = $path;
		}
	}

	# Clean up search path
	$searchpath =~ s{/+}{/};
	$searchpath =~ s{/[^/]*$}{} if length $searchpath > 1;

	my $pathmap = $bsc->{conf}{pathmap};
	if( $pathmap ) {
		$path =~ s{^/|/$}{}g;
		my @path = split '/', $path;
		for( my $i = @path; $i > 0; --$i ) {
			my $checkpath = '/'.join('/', @path[0..($i-1)] );
			if( $pathmap->{$checkpath} ) {
				$basedir = $pathmap->{$checkpath};
				$searchpath = @path > $i ? '/'.join('/',@path[$i..(@path-1)]) : '/';
				last;
			} elsif( exists $pathmap->{$checkpath} ) {
				return;
			}
		}
	}

	my $fh;
	my $keep_looking = 1;
	while( $keep_looking ) {
		my $no_find_time = $no_allowed_netids_cache{$searchpath};
		unless( $no_find_time and (time - $no_find_time < ALLOWED_NETIDS_CACHE_TIMEOUT ) ) {
			$fh = FileHandle->new("$basedir$searchpath/allowed.netids");
			$fh and last;
			$no_allowed_netids_cache{$searchpath} = time;
		}
		$keep_looking = $searchpath =~ s{/[^/]*$}{};
	}

	# Couldn't find an allowed_netids file
	$fh or return;

	return __parse_allowed_netids($fh,$bsc);
}

sub get_bluestem_client ($)
{
	my $r = shift;
	my $conf = $ENV{BLUESTEM_CLIENT_CONFIG} || '';
	$bsc_cache{$conf} ||= Bluestem::Client->new( conf => $conf );
	return $bsc_cache{$conf};
}

sub __remote_ip ($$)
{
	my($r,$bsc) = @_;
	my $ip = r_remote_ip($r);
	if( r_header_in($r,'X-Forwarded-For')
	 and $bsc->{conf}{proxy_forwarders}
	 and $bsc->{conf}{proxy_forwarders}{$ip}
	) {
		$ip = $ENV{HTTP_X_FORWARDED_FOR};
	}
	return $ip;
}

sub clear_bluestem_cookies ($)
{
        my $r = shift;
        my $cookiehdr = $r->headers_in->{'Cookie'} or return;
        my @cookies = grep {$_ !~ m/^\s*Bluestem_[\w.\-]*=/} split /;/, $cookiehdr;
        $r->headers_in->set( 'Cookie' => join(';',@cookies) );
}

sub get_session ($$)
{
	my($r,$bsc) = @_;
	my($cacheid,$clientkey,$paservers) = read_session_cookies($r,$bsc);
	my $session = $bsc->__get_session($cacheid,$clientkey,__remote_ip($r,$bsc))
		if $cacheid;
	return $session unless wantarray;
	return($session,$cacheid,$clientkey,$paservers);
}

sub authen_handler ($)
{
	my $r = shift;
	my $bsc = get_bluestem_client($r);

	# Always ignore for bluestem cgi or images
	my $path = r_path($r);
	$path =~ s{/+}{/}g;
	my $cgipath = $bsc->{conf}{cgipath};
	$cgipath =~ s{/$}{};
	$path =~ m{^$cgipath/[^/]+$} and return;
	my $imgpath = $bsc->{conf}{imgpath};
	$imgpath =~ s{/$}{};
	$path =~ m{^$imgpath/[^/]+$} and return;

	my($session,$cacheid,$clientkey,$paservers) = get_session($r,$bsc);

	unless( $session ) {
		my $iserver = $bsc->choose_id_server( $paservers );
		($cacheid,$clientkey) = $bsc->create_cacheentry($iserver,__remote_ip($r,$bsc));
		my $sslurl = "https://".r_server_name($r).r_path($r);
		my $query_string = r_query_string($r);
		$sslurl .= "?$query_string" if $query_string;
		my $login_url = $bsc->login_url(
			cacheid => $cacheid,
			iserver => $iserver,
			message => r_perlvar($r,"BluestemMessage"),
			return_url => $sslurl,
		);
		my $cookiestr = $bsc->client_cookie($cacheid,$clientkey);
		r_add_err_header($r,"Set-cookie",$cookiestr);
		r_add_err_header($r,"Location",$login_url);
		return REDIRECT;
	}

	$bsc->update_cacheaccess( $cacheid );
	my $netid = $bsc->bluestem_id_short( $session->{netid} );
	r_user($r,$netid);
	
	while( my($k,$v) = each %{$session->{userinfo}} ) {
		if( UNIVERSAL::isa($v,'ARRAY') ) {
			$v = join ', ', @$v;
		} elsif( UNIVERSAL::isa($v,'HASH') ) {
			$v = join ', ', map {"$_=$v->{$_}"} keys %$v;
		}
		r_subprocess_env($r,"BSUSER_$k" => $v);
	}

	my $cookiehdr = r_header_in($r,'Cookie');
	my @cookies = grep {$_ !~ m/^\s*$bsc->{conf}{client_cookiename}=/} split /;/, $cookiehdr;

	my $tktauth_secret = $r->dir_config('TKTAuthSecret')
		|| $bsc->{conf}{tkt_auth_secret};
	if( $tktauth_secret ) {
		$have_authtkt or die "No Apache::AuthTkt module found!\n";
		my $tktauth_digest = $r->dir_config('TKTAuthDigest')
			|| $bsc->{conf}{tkt_auth_digest}
			|| 'MD5';
		my $tkt_cookiename = $r->dir_config('TKTAuthCookieName')
			|| $bsc->{conf}{tkt_auth_cookie}
			|| 'auth_tkt';
		my $at = Apache::AuthTkt->new(
			secret => $tktauth_secret,
			digest_type => $tktauth_digest,
		);
		my $ticket = $at->ticket( uid => $bsc->bluestem_id_short($session->{netid}), ip_addr => __remote_ip($r,$bsc) );
		push @cookies, "$tkt_cookiename=$ticket";
	}

        if( $bsc->{conf}{username_header} ) {
                r_set_header_in($r, $bsc->{conf}{username_header} => $session->{netid});
        }

	r_set_header_in($r,'Cookie' => join('; ',@cookies));

	return OK;
}

sub access_handler ($)
{
	my $r = shift;
	my $bsc = get_bluestem_client($r);
	my $path = r_path($r) || '';

        $r->is_initial_req or return DECLINED;

        # Always ignore allowed.netids for bluestem cgi or images
	my $cgipath = $bsc->{conf}{cgipath};
	$cgipath =~ s{/$}{};
	$path =~ m{^$cgipath/[^/]+$} and return DECLINED;
	my $imgpath = $bsc->{conf}{imgpath};
	$imgpath =~ s{/$}{};
	$path =~ m{^$imgpath/[^/]+$} and return DECLINED;

	# Ignore index files unless such a file actually exists
	my $filename = $r->filename();
	$filename =~ m{/(default|index)\.(html?|php)} && !-e $filename
		and return DECLINED;

	# Otherwise handle as bluestem
	my($session,$cacheid,$clientkey,$paservers) = get_session($r,$bsc);
	my $netid;
	if( $session ) {
		$bsc->update_cacheaccess( $cacheid );
		$netid = $bsc->bluestem_id_short( $session->{netid} );
		r_user($r,$netid);
		if( $session->{userinfo} ) {
			while( my($k,$v) = each %{$session->{userinfo}} ) {
				if( UNIVERSAL::isa($v,'ARRAY') ) {
					$v = join ', ', @$v;
				} elsif( UNIVERSAL::isa($v,'HASH') ) {
					$v = join ', ', map {"$_=$v->{$_}"} keys %$v;
				}
				r_subprocess_env($r,"BSUSER_$k" => $v);
			}
		}
	}

	clear_bluestem_cookies($r);

	$path =~ m{/allowed.netids$} 
		and return FORBIDDEN;

	my $allowed_netids = __get_allowed_netids($r,$bsc)
		or return DECLINED;

	# Permit based on ips
	my $clientip = __remote_ip($r,$bsc);
	for my $allowed_ip ( @{$allowed_netids->{ip}} ) {
		if( $allowed_ip =~ s/\*$// ) {
			return OK if
				substr($clientip, 0, length $allowed_ip) eq $allowed_ip;
		} else {
			return OK if $clientip eq $allowed_ip;
		}
	}

	# Permit based on hostname
	my $clienthost = r_remote_host($r);
	# If apache has hostname lookups tured off, have to do it here
	$clienthost ||= gethostbyaddr(inet_aton($clientip),AF_INET);
	for my $allowed_host ( @{$allowed_netids->{host}} ) {
		if( $allowed_host =~ s/^\*// ) {
			return OK if
				substr($clienthost, -length $allowed_host) eq $allowed_host;
		} else {
			return OK if $clienthost eq $allowed_host;
		}
	}

	# For anything else we need to invoke bluestem.

	# Client may already be authenticated, but not using an SSL request
	my $port = r_local_port($r);
	if( $port == 80 ) {
		my $sslurl = "https://".r_server_name($r).$path;
		my $query_string = r_query_string($r);
		$sslurl .= "?$query_string" if $query_string;
		r_add_err_header($r,"Location",$sslurl);
		return REDIRECT;
	}

	# If no session was found, the user needs to authenticate.  Do redirect.
	unless( $session ) {
		my $iserver = $bsc->choose_id_server( $paservers );
		($cacheid,$clientkey) = $bsc->create_cacheentry($iserver,$clientip);
		my $sslurl = "https://".r_server_name($r).$path;
		my $query_string = r_query_string($r);
		$sslurl .= "?$query_string" if $query_string;
		my $login_url = $bsc->login_url(
			cacheid => $cacheid,
			iserver => $iserver,
			message => r_perlvar($r,"BluestemMessage"),
			return_url => $sslurl,
		);
		my $cookiestr = $bsc->client_cookie($cacheid,$clientkey);
		r_add_err_header($r,"Set-cookie",$cookiestr);
		r_add_err_header($r,"Location",$login_url);
		return REDIRECT;
	}

	# Check if netid is allowed
	if( $allowed_netids->{netid}{$netid} ) {
		return OK;
	}

	# Check if allowed by group
	my($domain) = $netid =~ m/(@.*)/;
	$domain ||= '';
        if( $session->{userinfo}
		and $session->{userinfo}{AdGroups}
		and grep { $allowed_netids->{group}{$_} } split ';', $session->{userinfo}{AdGroups}
        ){
		return OK;
        }
	if( $allowed_netids->{group}{"all$domain"} ) {
		return OK;
	}
	if( $netid =~ m{^(.*)/(.*)$} ) {
		my $authmeth = $2;
		return OK if $allowed_netids->{group}{"all/$authmeth"};
	}

	my $contact = uri_escape($allowed_netids->{contact} || '');
	$r->custom_response(FORBIDDEN,"$bsc->{conf}{cgipath}/forbidden.cgi?contact=$contact&path=$path");
	return FORBIDDEN;
}

1;
